<?php
/**
 * The plugin logging class.
 *
 * @since      	1.0
 */
namespace dodebug;

defined( 'WPINC' ) || exit;

class Log extends Instance
{
	protected static $_instance;
	private static $log_path;
	private static $_prefix;

	private static $_ignore_filters;
	private static $_ignore_part_filters;

	const TYPE_CLEAR_LOG = 'clear_log';

	/**
	 * Log class Confructor
	 *
	 * NOTE: in this process, until last step ( define const debug=true ), any usage to WP filter will not be logged to prevent infinite loop with log_filters()
	 *
	 * @since 1.0
	 * @access public
	 */
	protected function __construct()
	{
		self::$log_path = WP_CONTENT_DIR . '/debug.log';

		! defined( 'DODEBUG_LOG_TAG' ) && define( 'DODEBUG_LOG_TAG', get_current_blog_id() );

		if ( Conf::val( 'adv' ) ) {
			! defined( 'debug2' ) && define( 'debug2', true );
		}
	}

	/**
	 * get the log file path
	 *
	 * @since 1.0
	 * @access public
	 */
	public function filepath()
	{
		return self::$log_path;
	}

	/**
	 * Enable debug log
	 *
	 * @since 1.0
	 * @access public
	 */
	public static function init()
	{
		$admin_ip = Conf::val( 'admin_ip' );
		if ( $admin_ip ) {
			if ( ! IP::get_instance()->maybe_hit_rule( $admin_ip ) ) {
				define( 'DODEBUG_LOG_BYPASS_NOTADMIN', true );
				return;
			}
		}

		/**
		 * Check if hit URI includes/excludes
		 * This is after DODEBUG_LOG_BYPASS_NOTADMIN to make `log_purge()` still work
		 */
		$list = Conf::val( 'uri_inc' );
		if ( $list ) {
			$result = Util::str_hit_array( $_SERVER[ 'REQUEST_URI' ], $list );
			if ( ! $result ) {
				return;
			}
		}

		$list = Conf::val( 'uri_exc' );
		if ( $list ) {
			$result = Util::str_hit_array( $_SERVER[ 'REQUEST_URI' ], $list );
			if ( $result ) {
				return;
			}
		}

		if ( ! defined( 'debug' ) ) { // If not initialized, do it now
			self::get_instance()->_init_request();
			define( 'debug', true );
		}

		// Check if hook filters
		if ( Conf::val( 'log_filter' ) ) {
			self::$_ignore_filters = Conf::val( 'filter_exc' );
			self::$_ignore_part_filters = Conf::val( 'filter_partial_exc' );

			add_action( 'all', __CLASS__ . '::log_filters' );
		}
	}

	/**
	 * Create the initial log messages with the request parameters.
	 *
	 * @since 1.0
	 * @access private
	 */
	private function _init_request( $log_file = null )
	{
		add_action( 'shutdown', array( $this, 'shutdown' ), 99999 ) ;

		if ( ! $log_file ) {
			$log_file = self::$log_path;
		}

		// Check log file size
		$log_file_size = Conf::val( 'log_size' );
		if ( file_exists( $log_file ) && filesize( $log_file ) > $log_file_size * 1000000 ) {
			f::save( $log_file, '' );
		}

		// For more than 2s's requests, add more break
		if ( file_exists( $log_file ) && time() - filemtime( $log_file ) > 2 ) {
			f::append( $log_file, "\n\n\n\n" );
		}

		if ( PHP_SAPI == 'cli' ) {
			return;
		}

		$servervars = array(
			'Query String' => '',
			'HTTP_ACCEPT' => '',
			'HTTP_USER_AGENT' => '',
			'HTTP_ACCEPT_ENCODING' => '',
			'HTTP_COOKIE' => '',
			'X-LSCACHE' => '',
			'LSCACHE_VARY_COOKIE' => '',
			'LSCACHE_VARY_VALUE' => '',
			'ESI_CONTENT_TYPE' => '',
		);
		$server = array_merge( $servervars, $_SERVER );
		$params = array();

		if ( isset( $_SERVER[ 'HTTPS' ] ) && $_SERVER[ 'HTTPS' ] == 'on' ) {
			$server[ 'SERVER_PROTOCOL' ] .= ' (HTTPS) ';
		}

		$param = sprintf( '💓 ------%s %s %s', $server[ 'REQUEST_METHOD' ], $server[ 'SERVER_PROTOCOL' ], strtok( $server[ 'REQUEST_URI' ], '?' ) );

		$qs = ! empty( $server[ 'QUERY_STRING' ] ) ? sanitize_text_field( $server[ 'QUERY_STRING' ] ) : '';

		$params[] = $param;
		$params[] = 'Query String: ' . $qs;

		if ( ! empty( $_SERVER[ 'HTTP_REFERER' ] ) ) {
			$params[] = 'HTTP_REFERER: ' . $server[ 'HTTP_REFERER' ];
		}

		if ( defined( 'debug2' ) ) {
			$params[] = 'User Agent: ' . $server[ 'HTTP_USER_AGENT' ];
			$params[] = 'Accept: ' . $server[ 'HTTP_ACCEPT' ];
			$params[] = 'Accept Encoding: ' . $server[ 'HTTP_ACCEPT_ENCODING' ];
		}
		if ( Conf::val( 'log_cookie' ) ) {
			$params[] = 'Cookie: ' . $server[ 'HTTP_COOKIE' ];
		}
		if ( isset( $_COOKIE[ '_lscache_vary' ] ) ) {
			$params[] = 'Cookie _lscache_vary: ' . sanitize_textarea_field( $_COOKIE[ '_lscache_vary' ] );
		}
		if ( defined( 'debug2' ) ) {
			$params[] = 'X-LSCACHE: ' . ( ! empty( $server[ 'X-LSCACHE' ] ) ? 'true' : 'false' );
		}
		if( $server[ 'LSCACHE_VARY_COOKIE' ] ) {
			$params[] = 'LSCACHE_VARY_COOKIE: ' . $server[ 'LSCACHE_VARY_COOKIE' ];
		}
		if( $server[ 'LSCACHE_VARY_VALUE' ] ) {
			$params[] = 'LSCACHE_VARY_VALUE: ' . $server[ 'LSCACHE_VARY_VALUE' ];
		}
		if( $server[ 'ESI_CONTENT_TYPE' ] ) {
			$params[] = 'ESI_CONTENT_TYPE: ' . $server[ 'ESI_CONTENT_TYPE' ];
		}

		$request = array_map( __CLASS__ . '::format_message', $params );

		f::append( $log_file, $request );
	}

	/**
	 * Shutdown callback
	 *
	 * @since 1.0
	 * @access public
	 */
	public function shutdown()
	{
		self::debug( "End response\n--------------------------------------------------------------------------------\n" );
	}

	/**
	 * Log all filters and action hooks
	 *
	 * @since 1.0
	 * @access public
	 */
	public static function log_filters()
	{
		$action = current_filter();

		if ( self::$_ignore_filters && in_array( $action, self::$_ignore_filters ) ) {
			return;
		}

		if ( self::$_ignore_part_filters ) {
			foreach ( self::$_ignore_part_filters as $val ) {
				if ( stripos( $action, $val ) !== false ) {
					return;
				}
			}
		}

		self::debug( "===log filter: $action" );
	}

	/**
	 * Formats the log message with a consistent prefix.
	 *
	 * @since 1.0
	 * @access private
	 */
	private static function format_message( $msg )
	{
		// If call here without calling get_enabled() first, improve compatibility
		if ( ! defined( 'DODEBUG_LOG_TAG' ) ) {
			return $msg . "\n";
		}

		if ( ! isset( self::$_prefix ) ) {
			// address
			if ( PHP_SAPI == 'cli' ) {
				$addr = '=CLI=';
				if ( isset( $_SERVER[ 'USER' ] ) ) {
					$addr .= $_SERVER[ 'USER' ];
				}
				elseif ( $_SERVER[ 'HTTP_X_FORWARDED_FOR' ] ) {
					$addr .= $_SERVER[ 'HTTP_X_FORWARDED_FOR' ];
				}
			}
			else {
				$addr = $_SERVER[ 'REMOTE_ADDR' ] . ':' . $_SERVER[ 'REMOTE_PORT' ];
			}

			// Generate a unique string per request
			self::$_prefix = sprintf( " [%s %s %s] ", $addr, DODEBUG_LOG_TAG, s::rrand( 3 ) );
		}
		list( $usec, $sec ) = explode(' ', microtime() );

		$backtrace = debug_backtrace();
		$ori_called = 4;
		if ( empty( $backtrace[ 4 ][ 'class' ] ) && ! empty( $backtrace[ 4 ][ 'function' ] ) && $backtrace[ 4 ][ 'function' ] == 'debug2' ) {
			$ori_called = 5;
		}
		$namespace = ! empty( $backtrace[ $ori_called ][ 'class' ] ) ? $backtrace[ $ori_called ][ 'class' ] : '';
		if ( strpos( $namespace, 'dodebug\\Core' ) === 0 ) {
			$namespace = '';
		}
		elseif ( strpos( $namespace, 'dodebug\\' ) === 0 ) {
			$namespace = substr( $namespace, 7 );
		}

		return date( 'm/d/y H:i:s', $sec + DODEBUG_TIME_OFFSET ) . substr( $usec, 1, 4 ) . self::$_prefix . $namespace . ' ' . $msg . "\n";
	}

	/**
	 * Direct call to log a debug message.
	 *
	 * @since 1.0
	 * @access public
	 */
	public static function debug( $msg, $backtrace_limit = false )
	{
		if ( ! defined( 'debug' ) ) {
			return;
		}

		if ( $backtrace_limit !== false ) {
			if ( ! is_numeric( $backtrace_limit ) ) {
				$msg .= ' --- ' . var_export( $backtrace_limit, true );
				self::push( $msg );
				return;
			}

			self::push( $msg, $backtrace_limit + 1 );
			return;
		}

		self::push( $msg );
	}

	/**
	 * Direct call to log an advanced debug message.
	 *
	 * @since 1.0
	 * @access public
	 */
	public static function debug2( $msg, $backtrace_limit = false )
	{
		if ( ! defined( 'debug2' ) ) {
			return;
		}
		self::debug( $msg, $backtrace_limit );
	}

	/**
	 * Logs a debug message.
	 *
	 * @since 1.0
	 * @access private
	 */
	private static function push( $msg, $backtrace_limit = false )
	{
		if ( defined( 'DODEBUG_LOG_TERMINATED' ) ) {
			return;
		}

		// backtrace handler
		if ( defined( 'debug2' ) && $backtrace_limit !== false ) {
			$msg .= self::_backtrace_info( $backtrace_limit );
		}

		f::append( self::$log_path, self::format_message( $msg ) );
	}

	/**
	 * Backtrace info
	 *
	 * @since 1.0
	 */
	private static function _backtrace_info( $backtrace_limit )
	{
		$msg = '';

		$trace = version_compare( PHP_VERSION, '5.4.0', '<' ) ? debug_backtrace( DEBUG_BACKTRACE_IGNORE_ARGS ) : debug_backtrace( DEBUG_BACKTRACE_IGNORE_ARGS, $backtrace_limit + 3 );
		for ( $i=2; $i <= $backtrace_limit + 2; $i++ ) { // 0st => _backtrace_info(), 1st => push()
			if ( empty( $trace[ $i ][ 'class' ] ) ) {
				if ( empty( $trace[ $i ][ 'file' ] ) ) {
					break;
				}
				$log = "\n" . $trace[ $i ][ 'file' ];
			}
			else {
				if ( $trace[ $i ][ 'class' ] == 'Log' ) {
					continue;
				}

				$log = $trace[ $i ][ 'class' ] . $trace[ $i ][ 'type' ] . $trace[ $i ][ 'function' ] . '()';
			}
			if ( ! empty( $trace[ $i - 1 ][ 'line' ] ) ) {
				$log .= '@' . $trace[ $i-1 ][ 'line' ];
			}
			$msg .= " => $log";
		}

		return $msg;
	}

	/**
	 * Clear log file
	 *
	 * @since 1.0
	 * @access public
	 */
	public function clear_log( $del = false )
	{
		if ( $del ) {
			unlink( self::$log_path );
			! defined( 'DODEBUG_LOG_TERMINATED' ) && define( 'DODEBUG_LOG_TERMINATED', true );
			return;
		}

		f::save( self::$log_path, '' );
		f::save( WP_CONTENT_DIR . '/debug.purge.log', '' );
	}

	/**
	 * Handle all request actions from main cls
	 *
	 * @since  1.0
	 * @access public
	 */
	public static function handler()
	{
		$instance = self::get_instance();

		$type = Router::verify_type();

		switch ( $type ) {
			case self::TYPE_CLEAR_LOG :
				$instance->clear_log();
				break;

			default:
				break;
		}

		Admin::redirect();
	}
}
